from typing import Tuple
import numpy as np
import pandas as pd
from math import pow, sqrt
import statsmodels.api as sm
from functools import wraps

class Indicator:
    """
    计算时间序列的各项指标
    net_value 和 benchmark的index必须一致
    假设净值序列等间隔分布
    """

    def __init__(self, net_values: pd.Series, benchmark: pd.Series) -> None:
        self.net_values = net_values
        self.benchmark = benchmark.reindex(net_values.index)
        self.isvalid = self.check_valid()

    def check_valid(self):
        if self.nv_num<2:
            return False

        intervals = pd.Series(self.net_values.index).diff().dropna()
        if max(intervals)>pd.Timedelta("50d"):
            return False
        
        return True

    def check(func):
        '''
        return None or the result.
        '''
        @wraps(func)
        def wrapper(self,*args, **kwargs):
            if self.isvalid==False:
                return np.nan
            result = func(self,*args, **kwargs)
            return result
        return wrapper

    @property
    def _returns(self):
        return self.net_values.pct_change().dropna()

    @property
    def _logreturns(self):
        return (np.log(self.net_values) - np.log(self.net_values.shift(1))).dropna()
    
    @property
    def _returns_benchmark(self):
        return self.benchmark.pct_change().dropna()

    @property
    def _excess_return_arithmetic(self):
        return self._returns - self._returns_benchmark

    @property
    def _excess_return_geometric(self):
        return (1+self._returns)/(1+self._returns_benchmark) - 1
    


    @property
    def _start(self) -> pd.Timestamp:
        if self.nv_num == 0:
            return np.nan
        return self.net_values.index.min()

    @property
    def _end(self) -> pd.Timestamp:
        if self.nv_num == 0:
            return np.nan
        return self.net_values.index.max()

    @property
    @check
    def interval(self) -> float:
        """
        净值序列的横跨天数
        """
        start_dt = pd.to_datetime(self._start)
        end_dt = pd.to_datetime(self._end)
        days = (end_dt-start_dt).days
        return days

    @property
    @check
    def interval_year(self) -> float:
        """
        净值序列的横跨年数
        """
        return self.interval/365

    @property
    def nv_num(self):
        """
        净值样本数量
        """
        return len(self.net_values)

    @property
    @check
    def nv_freq(self) -> float:
        """
        净值数据频率：平均每两个净值之间的间隔天数
        周频数据应约为 7
        """
        if self.nv_num == 0:
            return 0 
        return self.interval/self.nv_num

    @property
    @check
    def nv_per_year(self) -> float:
        """
        平均每年有多少个净值数据（用于年化）
        """
        return self.nv_num/self.interval_year

    @property
    @check
    def total_return(self) -> float:
        """
        区间总收益
        """
        return self._calc_return()

    @property
    @check
    def total_benchmark_return(self) -> float:
        """
        业绩基准区间总收益
        """
        return self._calc_return(bm=True)

    @property
    @check
    def annulized_return(self) -> float:
        """区间年化收益"""
        return self._calc_return(annulized=True)

    @property
    @check
    def annulized_benchmark_return(self):
        """区间业绩基准年化收益"""
        return self._calc_return(annulized=True, bm=True)

    @property
    @check
    def annulized_excess_return(self):
        """区间简单超额年化收益"""
        return self.annulized_return-self.annulized_benchmark_return

    # @cache
    def _calc_return(self, annulized=False, bm=False):
        if not bm:
            p_series = self.net_values
        elif bm:
            p_series = self.benchmark

        if len(p_series) == 0:
            return 0

        s = p_series.iloc[0]
        e = p_series.iloc[-1]
        ret = e/s-1

        if not annulized:
            return ret

        elif annulized:
            years = self.interval_year
            if years == 0:
                return 0
            else:
                return pow(ret+1, 1/years) - 1

    @property
    @check
    def volatility(self) -> float:
        """区间波动率"""
        return self._returns.std(ddof=1)

    @property
    @check
    def annulized_volatility(self) -> float:
        """
        区间年化波动率，频率是nv_freq
        转化为年化波动率 公式:sqrt(days in year(365)/nv_freq)*vol = sqrt(days*num_nv/interval)*vol = sqrt(num_nv/year_interval)*vol
        """
        if self.nv_freq == 0:
            return 0
        return sqrt(365/self.nv_freq)*self.volatility

    @property
    @check
    def alpha(self) -> float:
        """区间年化alpha"""
        return self._calc_alphabeta[0]

    @property
    @check
    def beta(self) -> float:
        """区间beta"""
        return self._calc_alphabeta[1]

    @property
    def _calc_alphabeta(self) -> Tuple[float, float]:
        
        y = self._returns
        X = self._returns_benchmark
        
        if len(X) == 1:
            return 0,1
        
        X = sm.add_constant(X)
        mod = sm.OLS(y, X)
        res = mod.fit()
        # 区间alpha
        alpha = res.params.const
        # 年化alpha
        if alpha<-1:
            annulized_alpha = np.nan
        else:
            annulized_alpha = np.power(alpha+1, 365/self.nv_freq) - 1
        # beta
        beta = res.params.iloc[1]

        return annulized_alpha, beta

    @property
    @check
    def corr(self) -> float:
        """相关系数"""
        return self._returns.corr(self._returns_benchmark)

    # @cache
    def calc_drawdown(self, bm=False):
        if not bm:
            p_series = self.net_values
        elif bm:
            p_series = self.benchmark
        
        if len(p_series) == 0:
            return 0
        
        water_mark = p_series.cummax()
        drawdown = 1-p_series/water_mark
        return max(drawdown)

    @property
    @check
    def drawdown(self) -> float:
        """区间最大回撤"""
        return self.calc_drawdown()

    @property
    @check
    def drawdown_benchmark(self) -> float:
        """业绩基准区间最大回撤"""
        return self.calc_drawdown(bm=True)

    @property
    @check
    def relative_drawdown(self) -> float:
        """区间相对回撤"""
        return self.drawdown-self.drawdown_benchmark

    # @cache
    @check
    def sharpe(self, rf=0.015) -> float:
        """
        夏普比率
        parameters:
        rf:无风险收益率
        """
        if self.annulized_volatility == 0:
            return 0
        return (self.annulized_return-rf)/self.annulized_volatility

    @property
    @check
    def ir(self) -> float:
        """信息比率"""
        alpha, beta = self._calc_alphabeta
        exceed_return = self._returns - beta * self._returns_benchmark
        # 超额序列波动率
        std_er = exceed_return.std(ddof=1)
        # 年化超额波动率
        annulized_std_er = sqrt(365/self.nv_freq)*std_er
        # 信息比率
        if annulized_std_er == 0:
            return 0
        ir = alpha/annulized_std_er
        return ir

    @property
    @check
    def downside_deviation(self):
        ret = self._returns.copy()
        ret[ret > 0] = 0
        downside_variance = ret.pow(2).sum()/len(ret)
        downside_deviation = sqrt(downside_variance)
        return downside_deviation

    # @cache
    @check
    def sortino(self, mar=0.015) -> float:
        """
        索提诺比率
        -----------
        parameters:
        mar:minimal accepted return 要求最低回报率(目标回报率)

        note:
        下行方差 = 低于mar的return与mar之间的差值的平方/总样本数(自由度为0)
        下行标准差 = sqrt(下行方差)

        年化索提诺:(年化收益率-mar)/年化下行标准差
        """
        # TODO : mar应该对于周频数据生效，目前取0

        downside_deviation = self.downside_deviation
        if (downside_deviation == 0) or (self.nv_freq == 0):
            return 0 
        # 转化为年化下行波动率 公式：sqrt(days/avgdays_per_nv)*vol = sqrt(days*num_nv/interval)*vol = sqrt(num_nv/year_interval)*vol
        annulized_downside_deviation = sqrt(
            365/self.nv_freq)*downside_deviation

        if annulized_downside_deviation == 0:
            return 0
        return (self.annulized_return-mar)/annulized_downside_deviation



from typing import Dict, List
from .product import Product
import pandas as pd

class Manager:

    def __init__(self) -> None:
        self.products: Dict[str, Product] = {}
        self.name: str = ""

    def add_product(self, prod: Product) -> None:
        self.products[prod.name] = prod

    def remove_product(self, prod_name: str) -> None:
        if prod_name in self.products:
            self.products.pop(prod_name)

    @property
    def AUM(self) -> float:
        aum = 0
        for p in self.products.values():
            aum += p.size
        return aum

    @property
    def prods_year(self) -> Dict[str, float]:
        d = {}
        for prod_name, prod in self.products.items():
            d[prod_name] = prod.years

        return d

    @property
    def prod_gt_1y(self) -> List[str]:
        return [name for name, year in self.prods_year.items() if year >= 1.0]

    def performance(self,start=None,end=None,fields=None) -> pd.DataFrame:
        """
        计算管理人在一段时间内旗下各产品的业绩表现
        fields:所需的字段
        """
        perf_series_list = []
        for product_name,product in self.products.items():
            p_performance = product.statistics(start,end,fields)               
            p_performance["name"] = product_name
            perf_series_list.append(p_performance)
        performance_df = pd.concat(perf_series_list,axis=1).T
        del perf_series_list
        return performance_df
    
    def rolling_performance(self,start_dt=None, end_dt=None, window="365d", interval='90d', fields:List[str]=None)-> pd.DataFrame:
        rollingperf_df_list = []
        for product_name,product in self.products.items():
            p_performance = product.rolling_statistics(start_dt,end_dt,window=window,interval=interval,fields=fields)               
            p_performance["name"] = product_name
            rollingperf_df_list.append(p_performance)
        performance_df = pd.concat(rollingperf_df_list)
        del rollingperf_df_list
        return performance_df

    def segment_performance(self, end_date=None, segments=None, mode="all", fields:List[str]=None)-> pd.DataFrame:
        segmentperf_df_list = []
        for product_name,product in self.products.items():
            p_performance = product.segment_statistics(end_date=end_date, segments=segments, mode=mode, fields=fields)               
            p_performance["name"] = product_name
            segmentperf_df_list.append(p_performance)
        performance_df = pd.concat(segmentperf_df_list)
        del segmentperf_df_list
        return performance_df

    def score(self,start=None,end=None,fields=None):
        """
        计算管理人在一段时间内旗下各产品的得分情况
        fields:所需的字段
        """
        score_series_list = []
        for product_name,product in self.products.items():
            p_score = product.score(start,end,fields)
            p_score["name"] = product_name
            score_series_list.append(p_score)
        score_df = pd.concat(score_series_list,axis=1).T
        del score_series_list
        return score_df

    def rolling_score(self,start_dt=None, end_dt=None, window="365d", interval='90d', fields:List[str]=None):
        rollingscore_df_list = []
        for product_name,product in self.products.items():
            p_score = product.rolling_score(start_dt,end_dt,window=window,interval=interval,fields=fields)               
            p_score["name"] = product_name
            rollingscore_df_list.append(p_score)
        score_df = pd.concat(rollingscore_df_list)
        del rollingscore_df_list
        return score_df

    def segment_score(self, end_date=None, segments=None, mode="all", fields:List[str]=None)-> pd.DataFrame:
        segmentscore_df_list = []
        for product_name,product in self.products.items():
            p_score = product.segment_scores(end_date=end_date, segments=segments, mode=mode, fields=fields)               
            p_score["name"] = product_name
            segmentscore_df_list.append(p_score)
        score_df = pd.concat(segmentscore_df_list)
        del segmentscore_df_list
        return score_df

    @property
    def prod_netvalues(self):
        nv_list = [p.net_values for _,p in self.products.items()]
        index=[pname for pname,_ in self.products.items()]
        return pd.DataFrame(nv_list,index=index).T




import pandas as pd


def calc_adjusted_netvalues(netvalues: pd.Series, cum_netvalues: pd.Series):
    netvalues = netvalues.sort_index()
    cum_netvalues = cum_netvalues.sort_index()
    dividend = (cum_netvalues-netvalues).diff().fillna(0)
    adjust_factor = (netvalues+dividend)/netvalues
    cum_adjust_factor = adjust_factor.cumprod()
    adj_cum_nv = cum_adjust_factor*netvalues
    adj_cum_nv.name = "分红复权累计净值"
    return pd.to_numeric(adj_cum_nv)


def truncate_timeseries(ts: pd.Series, start_time: pd.Timestamp, end_time: pd.Timestamp):
    return ts[(ts.index >= start_time) & (ts.index <= end_time)]



import pandas as pd
from pandas.core.dtypes.common import is_datetime64_any_dtype, is_numeric_dtype


class TimeSerise:
    """
    金融时间序列描述器
    pd.Series
        index:时间 dtype:pd.Datetimeindex
        values:价格 dtype:np.float
    """

    def __set_name__(self, owner, name):
        self.public_name = name
        self.private_name = '_' + name

    def __get__(self, obj, objtype):
        return getattr(obj, self.private_name, None)

    def __set__(self, obj, value):
        if value is None:
            return

        # value不是符合要求的pd.series
        try:
            value = value.dropna().sort_index()
            assert isinstance(value, pd.Series)
            assert is_datetime64_any_dtype(value.index)
            assert is_numeric_dtype(value)
            assert len(value) > 0
        except:
            return

        setattr(obj, self.private_name, value)
        obj._validate()

    def __delete__(self, obj):
        delattr(obj, self.private_name)


from typing import Dict
from .indicators import Indicator
from warnings import warn
import numpy as np
from functools import wraps

class Score:
    """
    打分标准
    """

    def __init__(self, indicator: Indicator, strategy: str) -> None:
        self.indcator = indicator
        self.strategy = strategy
    
    def check(func):
        '''
        return None or the result.
        '''
        @wraps(func)
        def wrapper(self,*args, **kwargs):
            if self.indcator.isvalid==False:
                return np.nan
            result = func(self,*args, **kwargs)
            return result
        return wrapper

    def score(self, value, threshold, ascending=False, maxscore=None, interval=1):
        """打分器"""
        if maxscore is None:
            maxscore = len(threshold)

        for t in threshold:
            # threshold降序(越大越好) 升序(越小越好)
            if ((not ascending) and (value >= t)) or (ascending and value <= t):
                return maxscore-threshold.index(t)*interval
        return 0.0

    @property
    @check
    def return_(self) -> float:
        """绝对收益得分"""
        thresholds: Dict = {"股票多头": [0.25, 0.2, 0.15, 0.1, 0.05,0.0],
                            "股票多空": [0.25, 0.2, 0.15, 0.1, 0.05,0.0],

                            "市场中性": [0.15, 0.12, 0.08, 0.05, 0.0],
                            "债券策略": [0.15, 0.12, 0.08, 0.05, 0.0],
                            "套利策略": [0.15, 0.12, 0.08, 0.05, 0.0],

                            "CTA策略": [0.2, 0.16, 0.12, 0.08, 0.0],
                            "复合策略": [0.2, 0.16, 0.12, 0.08, 0.0],
                            "其他": [0.2, 0.16, 0.12, 0.08, 0.0],
                            }

        threshold = thresholds.get(self.strategy, thresholds["股票多头"])
        ret = self.indcator.annulized_return

        return self.score(ret, threshold,maxscore=30,interval=5)

    @property
    @check    
    def excessreturn(self) -> float:
        """相对收益得分"""
        thresholds: Dict = {"股票多头": [0.2, 0.15, 0.1, 0.05, 0.0],
                            "股票多空": [0.2, 0.15, 0.1, 0.05, 0.0],

                            "市场中性": [0.1, 0.06, 0.03, 0.0],
                            "债券策略": [0.1, 0.06, 0.03, 0.0],
                            "套利策略": [0.1, 0.06, 0.03, 0.0],

                            "CTA策略": [0.15, 0.1, 0.05, 0.0],
                            "复合策略": [0.15, 0.1, 0.05, 0.0],
                            "其他": [0.15, 0.1, 0.05, 0.0],
                            }

        threshold = thresholds.get(self.strategy, thresholds["股票多头"])
        
        excess_return = self.indcator.annulized_excess_return
        if self.strategy == "股票多头":
            return self.score(excess_return, threshold,maxscore=20,interval=4)
        else:
            return self.score(excess_return, threshold,maxscore=20,interval=5)

    @property
    @check    
    def relative_drawdown(self) -> float:
        """相对回撤得分"""
        thresholds: Dict = {"股票多头": [-0.05, 0.0, 0.05, 0.08, 0.12],
                            "股票多空": [-0.05, 0.0, 0.05, 0.08, 0.12],

                            "市场中性": [0.0, 0.02, 0.04, 0.06, 0.08],
                            "债券策略": [0.0, 0.02, 0.04, 0.06, 0.08],
                            "套利策略": [0.0, 0.02, 0.04, 0.06, 0.08],

                            "CTA策略": [-0.03, 0.0, 0.03, 0.06, 0.1],
                            "复合策略": [-0.03, 0.0, 0.03, 0.06, 0.1],
                            "其他": [-0.03, 0.0, 0.03, 0.06, 0.1],
                            }
        threshold = thresholds.get(self.strategy, thresholds["股票多头"])

        relative_dd = self.indcator.relative_drawdown
        return self.score(relative_dd, threshold, ascending=True,maxscore=20,interval=4)

    @property
    @check    
    def sharpe(self) -> float:
        """夏普比率得分"""
        threshold = [2, 1.5, 1.2, 1, 0]
        sharpe = self.indcator.sharpe()

        return self.score(sharpe, threshold,maxscore=10,interval=2)

    @property
    @check    
    def sortino(self) -> float:
        """索提诺比率得分"""
        threshold = [2, 1.5, 1.2, 1, 0]
        sortino = self.indcator.sortino()

        return self.score(sortino, threshold,maxscore=10,interval=2)

    @property
    @check    
    def ir(self) -> float:
        """信息比率得分"""
        threshold = [1, 0.75, 0.5, 0.25, 0]
        ir = self.indcator.ir
        
        return self.score(ir, threshold,maxscore=10,interval=2)



from datetime import datetime, timedelta
from tracemalloc import start
from typing import List, Tuple
import pandas as pd
from pandas.core.dtypes.common import is_datetime64_any_dtype, is_numeric_dtype
from warnings import warn
from .indicators import Indicator
from .score import Score
from .object import TimeSerise
# from functools import cached_property, cache
from .util import truncate_timeseries

class Product:

    net_values = TimeSerise()
    benchmark = TimeSerise()

    def __init__(
        self,
        net_values: pd.Series = None,       # 产品净值序列
        benchmark: pd.Series = None,        # 业绩基准净值序列
        name: str = None,                   # 产品名称
        foundation_date: datetime = None,   # 产品创立时间
        manager: str = None,                # 基金经理
        code=None,                          # 产品创立时间
        size=None,                          # 产品规模
        strategy_type=None,                 # 策略类型
    ) -> None:
        self.name = name
        self.net_values = net_values
        self.benchmark = benchmark


        if (name is None) & (self.net_values is not None):
            self.name = self.net_values.name

        self.foundation_date = foundation_date
        if (foundation_date is None) & (self.net_values is not None):
            self.foundation_date = self.start

        self.manager = manager
        self.code = code
        self.size = size
        self.strategy_type = strategy_type

    def _validate(self):
        """
        检查业绩基准是否满足要求
        检查净值的日期是否与业绩基准对齐
        """
        if (self.net_values is not None) & (self.benchmark is not None):
            # benchmark是否需要更新
            if (self.benchmark.index.min() > self.start) | (self.benchmark.index.max() < self.end):
                del self.benchmark
                warn(f"{self.name}的benchmark的值不足以覆盖产品净值区间,已移除benchmark,请重新传入")


            # detect abnormal index
            abnormal_index = self.net_values.index[~self.net_values.index.isin(
                self.benchmark.index)]
            if len(abnormal_index) > 0:
                correction = {}
                for error_index in abnormal_index:
                    correct_index = self.benchmark.index.asof(error_index)
                    correction[error_index] = correct_index

                self.net_values.rename(correction, inplace=True)
                self.net_values = self.net_values[~self.net_values.index.duplicated(
                    keep='first')]

    @property
    def _tradedays(self):
        return self.benchmark.index

    @property
    def start(self):
        return self.net_values.index.min()

    @property
    def end(self):
        return self.net_values.index.max()

    @property
    def years(self):
        """产品存续总时长"""
        return (self.end-self.start).days/365

    # @cache
    def _truncate_product(self, start_dt:pd.Series, end_dt:pd.Series) -> Tuple[pd.Series,pd.Series]:
        sub_nv = truncate_timeseries(self.net_values,start_time=start_dt,end_time=end_dt)
        sub_bm = truncate_timeseries(self.benchmark,start_time=start_dt,end_time=end_dt)
        return (sub_nv,sub_bm)

    # @cache
    def _get_indicator(self, start_dt:pd.Series, end_dt:pd.Series) -> Indicator:
        """返回产品在某一区间内的指标"""
        nv, bm = self._truncate_product(start_dt, end_dt)
        return Indicator(nv, bm)

    # @cache
    def _get_score(self, start_dt:pd.Series, end_dt:pd.Series) -> Score:
        """返回产品在某一区间内的打分"""
        period_indicator = self._get_indicator(start_dt, end_dt)
        return Score(period_indicator, self.strategy_type)

    def _get_rollingback_periods(self, interval:str, window_size:str, wacth_point:str) -> List[Tuple[pd.Timestamp, pd.Timestamp]]:
        """
        滚动切分产品时间段
        parameters:
            interval:每次向前回溯的间隔
            window_size:每个评价期的长度,最后一个评价期会合并超出部分的净值
            wacth_point:回溯起始时间点
        return:
            List[Tuple(start_point,end_point)]
        """

        # 计算评价周期数
        interval = pd.to_timedelta(interval)
        window_size = pd.to_timedelta(window_size)
        total_duration = pd.to_timedelta(wacth_point-self.start)
        if total_duration<window_size:
            # 产品总时长不足一个评价期
            num_windows = 1
        else: 
            num_windows = int((total_duration-window_size)/interval)+1

        periods = []
        for i in range(num_windows):
            end_point = wacth_point-i*interval
            if i == num_windows-1:
                # 最后一个周期
                start_point = self.start
            else:
                start_point = end_point - window_size
            periods.append((start_point, end_point))

        return periods

    def _get_periods(self, mode:str, segments:List[int], watch_point:str) -> List[Tuple[pd.Timestamp, pd.Timestamp]]:
        """
        指定回溯的起始点和时段切分设置，获取分析需要的周期
        parameters:
            segments:切分时间点设置,
            watch_point:回溯的起始日期，
            mode:"all":获取自不同起点至今的时间段
                :"range":获取不同起点之间的时间段
        """

        starts_dt = [
            watch_point-timedelta(d) for d in segments if watch_point-timedelta(d) >= self.start]
        if mode == "all":
            ends_dt = [watch_point for _ in range(len(starts_dt))]
        elif mode == "range":
            ends_dt = starts_dt.copy()
        starts_dt.append(self.start)
        ends_dt.insert(0, watch_point)

        return list(zip(starts_dt, ends_dt))

    # 指标

    def statistics(self, start_dt=None, end_dt=None, fields:List[str]=None) -> pd.Series:
        """
        产品区间指标
        fields:
            total_return
            annulized_return
        """
        if fields is None:
            fields = ["total_return","total_benchmark_return","annulized_return","annulized_excess_return","drawdown","relative_drawdown","volatility","annulized_volatility","alpha","beta","sharpe","ir","sortino"]
        if isinstance(fields,str):
            fields = [fields]
        
        if start_dt is None:
            start_dt = self.start
        else:
            start_dt = pd.Timestamp(start_dt)

        if end_dt is None:
            end_dt = self.end
        else:
            start_dt = pd.Timestamp(start_dt)

        ind = self._get_indicator(start_dt, end_dt)

        temp_values = []
        for field in fields:
            value = getattr(ind,field)
            if callable(value):
                value = value()
            temp_values.append(value)
        temp_values.append(getattr(ind,"_start"))
        temp_values.append(getattr(ind,"_end"))
        temp_values.append(getattr(ind,"nv_num"))

        temp_values.extend([start_dt,end_dt])
        index = fields.copy()
        index.extend(["nv_start","nv_end","nv_num","start_date","end_date"])

        return pd.Series(data=temp_values,index=index)



    def rolling_statistics(self, start_dt=None, end_dt=None, window="365d", interval='1d', fields:List[str]=None) -> pd.DataFrame:
        """
        产品滚动业绩指标
        parameters:
            start_dt:滚动起始日
            end_dt:滚动终止日
            window:窗口期大小
            intervals:回滚间隔
        return:
            pd.Dataframe:
                index:datetimeindex
                values:区间收益
        """
        if start_dt == None:
            start_dt = self.start
        else:
            start_dt = pd.Timestamp(start_dt)

        if end_dt == None:
            end_dt = self.end
        else:
            end_dt = pd.to_datetime(end_dt)

        if start_dt > end_dt:
            raise ValueError("滚动起始日不能晚于终止日")

        if fields is None:
            fields = ["total_return","total_benchmark_return","annulized_return","annulized_excess_return","drawdown","relative_drawdown","volatility","annulized_volatility","alpha","beta","sharpe","ir","sortino"]
        if isinstance(fields,str):
            fields = [fields]
        
        periods = self._get_rollingback_periods(
            interval=interval,
            window_size=window,
            wacth_point=end_dt
        )

        temp = []
        for period in periods:
            if period[0] < start_dt:
                continue
            
            stats = self.statistics(period[0],period[1],fields)
            temp.append(stats)
        return pd.DataFrame(temp).set_index(["start_date","end_date"])
    
    def segment_statistics(self,end_date=None, segments=None, mode="all", fields:List[str]=None) -> pd.DataFrame:
        """
        指定终止日,按segments的时间长度进行回溯
        parameters:
            segments:切分时间点设置,默认[180, 365, 365*2, 365*3](半年,一年,2年,3年)
            end_date:回溯的起始日期，默认到最新净值(最后一个净值),
            mode:"all":获取自不同起点至今的时间段
                :"range":获取不同起点之间的时间段
            fields:需要的字段
        """
        if end_date == None:
            end_date = self.end
        else:
            end_date = pd.to_datetime(end_date)
        
        if segments is None:
            days = [180, 365, 365*2, 365*3]
        else:
            days = segments        
        if isinstance(days,int):
            days = [days]
        
        if fields is None:
            fields = ["total_return","total_benchmark_return","annulized_return","annulized_excess_return","drawdown","relative_drawdown","volatility","annulized_volatility","alpha","beta","sharpe","ir","sortino"]
        if isinstance(fields,str):
            fields = [fields]

        periods = self._get_periods(
            mode=mode,
            segments=days,
            watch_point=end_date
        )

        temp = []
        for period in periods:
            if period[0] < self.start:
                continue
            
            stats = self.statistics(period[0],period[1],fields)
            temp.append(stats)
        return pd.DataFrame(temp).set_index(["start_date","end_date"])


    def score(self,start_dt=None,end_dt=None,fields=None) -> pd.Series:
        if start_dt == None:
            start_dt = self.start
        else:
            start_dt = pd.Timestamp(start_dt)

        if end_dt == None:
            end_dt = self.end
        else:
            end_dt = pd.to_datetime(end_dt)
        
        if fields is None:
            fields = ["return_","excessreturn","relative_drawdown","sharpe","sortino","ir"]
        if isinstance(fields,str):
            fields = [fields]
        
        score = self._get_score(start_dt,end_dt)

        temp_values = []
        for field in fields:
            value = getattr(score,field)
            if callable(value):
                value = value()
            temp_values.append(value)
        
        temp_values.extend([start_dt,end_dt])
        index = fields.copy()
        index.extend(["start_date","end_date"])
        
        return pd.Series(data=temp_values,index=index)

    def rolling_score(self, start_dt=None, end_dt=None, window="365d", interval='1d', fields:List[str]=None) -> pd.DataFrame:

        if start_dt == None:
            start_dt = self.start
        else:
            start_dt = pd.Timestamp(start_dt)

        if end_dt == None:
            end_dt = self.end
        else:
            end_dt = pd.to_datetime(end_dt)

        if start_dt > end_dt:
            raise ValueError("滚动起始日不能晚于终止日")

        if fields is None:
            fields = ["return_","excessreturn","relative_drawdown","sharpe","sortino","ir"]
        if isinstance(fields,str):
            fields = [fields]
        
        periods = self._get_rollingback_periods(
            interval=interval,
            window_size=window,
            wacth_point=end_dt
        )

        temp = []
        for period in periods:
            if period[0] < start_dt:
                continue
            
            scores = self.score(period[0],period[1],fields)
            temp.append(scores)
        return pd.DataFrame(temp).set_index(["start_date","end_date"])

    def segment_scores(self,end_date=None, segments=None, mode="all", fields:List[str]=None) -> pd.DataFrame:
        """
        指定终止日,按segments的时间长度进行回溯
        parameters:
            segments:切分时间点设置,默认[180, 365, 365*2, 365*3](半年,一年,2年,3年)
            end_date:回溯的起始日期，默认到最新净值(最后一个净值),
            mode:"all":获取自不同起点至今的时间段
                :"range":获取不同起点之间的时间段
            fields:需要的字段
        """
        if end_date == None:
            end_date = self.end
        else:
            end_date = pd.to_datetime(end_date)
        
        if segments is None:
            days = [180, 365, 365*2, 365*3]
        else:
            days = segments        
        if isinstance(days,int):
            days = [days]
        
        if fields is None:
            fields = ["return_","excessreturn","relative_drawdown","sharpe","sortino","ir"]
        if isinstance(fields,str):
            fields = [fields]

        periods = self._get_periods(
            mode=mode,
            segments=days,
            watch_point=end_date
        )

        temp = []
        for period in periods:
            if period[0] < self.start:
                continue
            
            stats = self.score(period[0],period[1],fields)
            temp.append(stats)
        return pd.DataFrame(temp).set_index(["start_date","end_date"])
